@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #ifndef libuser_error_h
@@ -25,7 +25,7 @@
 
 G_BEGIN_DECLS
 
-typedef enum lu_status {
+enum lu_status {
 	/* Non-fatal. */
 	lu_success = 0,
 	lu_warning_config_disabled,
@@ -62,16 +62,50 @@ typedef enum lu_status {
 	lu_error_unlock_empty,
 
 	/* Since 0.56 */
-	lu_error_invalid_attribute_value
-} lu_status_t;
+	lu_error_invalid_attribute_value,
 
-typedef struct lu_error {
+	/* Since 0.57 */
+	lu_error_invalid_module_combination,
+
+	/* Since 0.60 */
+	lu_error_homedir_not_owned,
+};
+#ifndef __GTK_DOC_IGNORE__
+#ifndef LU_DISABLE_DEPRECATED
+typedef enum lu_status lu_status_t;
+#endif
+#endif
+
+/**
+ * lu_error:
+ *
+ * Error and status information.
+ */
+/* gtk-doc is dumb. */
+struct lu_error;
+struct lu_error {
 	enum lu_status code;
 	char *string;
-} lu_error_t;
+};
+#ifndef LU_DISABLE_DEPRECATED
+/**
+ * lu_error_t:
+ *
+ * An alias for struct #lu_error.
+ * Deprecated: 0.57.3: Use struct #lu_error directly.
+ */
+typedef struct lu_error lu_error_t;
+#endif
 
-/* Checks that a passed-in error pointer is not already populated, and calls
-   abort() if it is. */
+/**
+ * LU_ERROR_CHECK:
+ * @err_p_p: A pointer to a struct #lu_error * which will be checked.
+ *
+ * Checks that the given pointer to a pointer to a struct does not already
+ * point to a valid #lu_error structure, and calls abort() on failure.  This
+ * macro is used by many internal functions to check that an error has not
+ * already occurred when they are invoked.
+ */
 #define LU_ERROR_CHECK(err_p_p) \
 do { \
 	struct lu_error **__err = (err_p_p); \
@@ -86,14 +120,13 @@ do { \
 	} \
 } while(0)
 
-/* Functions for allocating and freeing error objects. */
-void lu_error_new(lu_error_t **error, lu_status_t code,
+void lu_error_new(struct lu_error **error, enum lu_status code,
 		  const char *fmt, ...) G_GNUC_PRINTF(3, 4);
-const char *lu_strerror(lu_error_t *error);
-gboolean lu_error_is_success(lu_status_t status);
-gboolean lu_error_is_warning(lu_status_t status);
-gboolean lu_error_is_error(lu_status_t status);
-void lu_error_free(lu_error_t **error);
+const char *lu_strerror(struct lu_error *error);
+gboolean lu_error_is_success(enum lu_status status);
+gboolean lu_error_is_warning(enum lu_status status);
+gboolean lu_error_is_error(enum lu_status status);
+void lu_error_free(struct lu_error **error);
 
 G_END_DECLS
 
