@@ -13,7 +13,7 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -213,15 +213,24 @@ connect_server(struct lu_ldap_context *context, struct lu_error **error)
 		return NULL;
 	}
 
-	/* Try to start TLS. */
-	ret = ldap_start_tls_s(ldap, NULL, NULL);
-	/* Note that TLS is not required for ldap:// URLs (unlike simple server
-	   names). */
-	if (ret != LDAP_SUCCESS && start_tls) {
-		lu_error_new(error, lu_error_init,
-			     _("could not negotiate TLS with LDAP server"));
-		close_server(ldap);
-		return NULL;
+	/* Skip STARTTLS for ldapi: Even if the server is set up for TLS in
+	   general, NSS doesn't support TLS over AF_UNIX; the STARTTLS
+	   operation would be accepted, but ldap_start_tls_s() would fail,
+	   keeping the connection broken to make authentication impossible.
+	   TLS on AF_UNIX does not make much sense anyway. */
+	if (strncmp(context->prompts[LU_LDAP_SERVER].value, "ldapi://", 8)
+	    != 0) {
+        	/* Try to start TLS. */
+	        ret = ldap_start_tls_s(ldap, NULL, NULL);
+		/* Note that TLS is not required for ldap:// URLs (unlike simple
+		   server names). */
+		if (ret != LDAP_SUCCESS && start_tls) {
+			lu_error_new(error, lu_error_init,
+				     _("could not negotiate TLS with LDAP "
+				       "server"));
+			close_server(ldap);
+			return NULL;
+		}
 	}
 
 	return ldap;
@@ -263,7 +272,7 @@ interact(LDAP *ld, unsigned flags, void *defs, void *xres)
 			break;
 		default:
 #ifdef DEBUG
-			g_print("Unhandled SASL Intreractive option `%d'.\n",
+			g_print("Unhandled SASL Intreractive option `%lu'.\n",
 				res[i].id);
 #endif
 			retval = LDAP_OTHER;
@@ -548,18 +557,8 @@ lu_ldap_lookup(struct lu_module *module,
 	ctx = module->module_context;
 
 	if (ent != NULL) {
-		GValueArray *array;
 		/* Try to use the dn the object already knows about. */
-
-		array = lu_ent_get(ent, DISTINGUISHED_NAME);
-		if (array != NULL) {
-			GValue *val;
-
-			val = g_value_array_get_nth(array, 0);
-			if (G_VALUE_HOLDS_STRING(val))
-				dn = g_value_get_string(val);
-		}
-
+		dn = lu_ent_get_first_string(ent, DISTINGUISHED_NAME);
 		if (dn == NULL)
 			/* Map the user or group name to an LDAP object name. */
 			dn = lu_ldap_ent_to_dn(module, namingAttr, name,
@@ -638,16 +637,12 @@ lu_ldap_lookup(struct lu_module *module,
 		if (ent_array != NULL)
 			ent = lu_ent_new_typed(type);
 		/* Set the distinguished name. */
-		memset(&value, 0, sizeof(value));
-		g_value_init(&value, G_TYPE_STRING);
 		p = ldap_get_dn(ctx->ldap, entry);
-		g_value_set_string(&value, p);
+		lu_ent_set_string_current(ent, DISTINGUISHED_NAME, p);
 		ldap_memfree(p);
-		lu_ent_clear_current(ent, DISTINGUISHED_NAME);
-		lu_ent_add_current(ent, DISTINGUISHED_NAME, &value);
-		g_value_unset(&value);
 
 		/* Read each of the attributes we asked for. */
+		memset(&value, 0, sizeof(value));
 		for (i = 0; attributes[i]; i++) {
 			BerValue **values;
 			const char *attr;
@@ -777,12 +772,7 @@ arrays_equal(GValueArray *a, GValueArray *b)
 {
 	GValue *aval, *bval;
 	size_t i, j;
-	if ((a != NULL) && (b == NULL)) {
-		return FALSE;
-	}
-	if ((a == NULL) && (b != NULL)) {
-		return FALSE;
-	}
+
 	for (i = 0; i < a->n_values; i++) {
 		aval = g_value_array_get_nth(a, i);
 		for (j = 0; j < b->n_values; j++) {
@@ -867,8 +857,8 @@ lu_ldap_needed_objectclasses(const char *dn, struct lu_ent *ent,
 	else
 		old_count = 0;
 
-	new_values = g_malloc(sizeof(*new_values) *
-			      (G_N_ELEMENTS(ldap_attribute_map) + 1 + 1));
+	new_values = g_malloc_n(G_N_ELEMENTS(ldap_attribute_map) + 1 + 1,
+				sizeof(*new_values));
 	new_count = 0;
 
 	/* Iterate over all of the attributes the object possesses. */
@@ -913,7 +903,7 @@ lu_ldap_needed_objectclasses(const char *dn, struct lu_ent *ent,
 		new_values[new_count] = bv;
 #ifdef DEBUG
 		g_print("Adding entity `%s' to class `%s'.\n", dn,
-			new_values[new_count]);
+			ldap_attribute_map[i].objectclass);
 #endif
 		new_count++;
 	}
@@ -976,8 +966,7 @@ get_ent_adds(const char *dn, struct lu_ent *ent)
 		GValue *value;
 		const GList *a;
 
-		mods = g_malloc0(sizeof(*mods)
-				 * (g_list_length(attrs) + 2 + 1));
+		mods = g_malloc0_n(g_list_length(attrs) + 2 + 1, sizeof(*mods));
 		mod_count = 0;
 		for (a = attrs; a != NULL; a = a->next) {
 			const char *attribute;
@@ -996,9 +985,8 @@ get_ent_adds(const char *dn, struct lu_ent *ent)
 			mod = g_malloc0(sizeof(*mod));
 			mod->mod_op = LDAP_MOD_ADD;
 			mod->mod_type = (char *)attribute;
-			mod->mod_values
-				= g_malloc0((vals->n_values + 1)
-					    * sizeof(*mod->mod_values));
+			mod->mod_values = g_malloc0_n(vals->n_values + 1,
+						      sizeof(*mod->mod_values));
 			for (i = 0; i < vals->n_values; i++) {
 				value = g_value_array_get_nth(vals, i);
 				mod->mod_values[i] = lu_value_strdup(value);
@@ -1013,8 +1001,8 @@ get_ent_adds(const char *dn, struct lu_ent *ent)
 			mod->mod_op = LDAP_MOD_ADD;
 			mod->mod_type = OBJECTCLASS;
 			mod->mod_values
-				= g_malloc0((ldap_count_values_len(classes)
-					     + 1) * sizeof(*mod->mod_values));
+				= g_malloc0_n(ldap_count_values_len(classes)
+					      + 1, sizeof(*mod->mod_values));
 			for (i = 0; classes[i] != NULL; i++)
 				mod->mod_values[i]
 					= g_strdup(classes[i]->bv_val);
@@ -1028,21 +1016,23 @@ get_ent_adds(const char *dn, struct lu_ent *ent)
 		    && lu_ent_get(ent, LU_COMMONNAME) == NULL) {
 			char *cn;
 
-			vals = lu_ent_get(ent, LU_GECOS);
-			if (vals != NULL) {
+			cn = lu_ent_get_first_value_strdup(ent, LU_GECOS);
+			if (cn != NULL) {
 				char *p;
 
-				value = g_value_array_get_nth(vals, 0);
-				cn = lu_value_strdup(value);
 				p = strchr(cn, ',');
 				if (p != NULL)
 					*p = 0;
-			} else {
-				vals = lu_ent_get(ent, LU_USERNAME);
+				/* Note that gecos may be empty, but
+				   commonName (as a DirectoryString) is not
+				   allowed to be empty. */
+			}
+			if (cn == NULL || *cn == 0) {
+				g_free(cn);
+				cn = lu_ent_get_first_value_strdup(ent,
+								   LU_USERNAME);
 				/* Guaranteed by lu_ldap_set() */
-				g_assert (vals != NULL);
-				value = g_value_array_get_nth(vals, 0);
-				cn = lu_value_strdup(value);
+				g_assert (cn != NULL);
 			}
 			mod = g_malloc0(sizeof(*mod));
 			mod->mod_op = LDAP_MOD_ADD;
@@ -1080,8 +1070,7 @@ get_ent_mods(struct lu_ent *ent, const char *namingAttr)
 		empty = g_value_array_new(0);
 		/* Allocate an array big enough to hold two LDAPMod structures
 		 * for each attribute, in case all of them need changing. */
-		mods = g_malloc0(sizeof(*mods) *
-				 ((2 * g_list_length(attrs)) + 1));
+		mods = g_malloc0_n(2 * g_list_length(attrs) + 1, sizeof(*mods));
 		mod_count = 0;
 		for (a = attrs; a != NULL; a = a->next) {
 			GValueArray *current, *pending, *additions, *deletions;
@@ -1129,8 +1118,8 @@ get_ent_mods(struct lu_ent *ent, const char *namingAttr)
 				mod->mod_op = LDAP_MOD_DELETE;
 				mod->mod_type = attribute;
 				mod->mod_values
-					= g_malloc0((deletions->n_values + 1)
-						    * sizeof(*mod->mod_values));
+					= g_malloc0_n(deletions->n_values + 1,
+						      sizeof(*mod->mod_values));
 				for (j = 0; j < deletions->n_values; j++) {
 					value = g_value_array_get_nth(deletions, j);
 					mod->mod_values[j]
@@ -1163,8 +1152,8 @@ get_ent_mods(struct lu_ent *ent, const char *namingAttr)
 				mod->mod_op = LDAP_MOD_ADD;
 				mod->mod_type = attribute;
 				mod->mod_values
-					= g_malloc0((additions->n_values + 1)
-						    * sizeof(*mod->mod_values));
+					= g_malloc0_n(additions->n_values + 1,
+						      sizeof(*mod->mod_values));
 				for (j = 0; j < additions->n_values; j++) {
 					value = g_value_array_get_nth(additions, j);
 					mod->mod_values[j]
@@ -1280,6 +1269,7 @@ lu_ldap_fudge_objectclasses(struct lu_ldap_context *ctx,
 		dump_mods(mods);
 #endif
 		err = ldap_modify_ext_s(ctx->ldap, dn, mods, NULL, NULL);
+		(void)err;
 #ifdef DEBUG
 		g_message("Fudged: `%s'.\n", ldap_err2string(err));
 #endif
@@ -1423,9 +1413,7 @@ static gboolean
 lu_ldap_del(struct lu_module *module, enum lu_entity_type type,
 	    struct lu_ent *ent, const char *branch, struct lu_error **error)
 {
-	GValueArray *name;
-	GValue *value;
-	char *name_string;
+	char *name;
 	const char *dn, *namingAttr;
 	int err;
 	gboolean ret = FALSE;
@@ -1446,7 +1434,7 @@ lu_ldap_del(struct lu_module *module, enum lu_entity_type type,
 		namingAttr = LU_GROUPNAME;
 	}
 
-	name = lu_ent_get(ent, namingAttr);
+	name = lu_ent_get_first_value_strdup(ent, namingAttr);
 	if (name == NULL) {
 		lu_error_new(error, lu_error_generic,
 			     _("object had no %s attribute"), namingAttr);
@@ -1454,10 +1442,8 @@ lu_ldap_del(struct lu_module *module, enum lu_entity_type type,
 	}
 
 	/* Convert the name to a distinguished name. */
-	value = g_value_array_get_nth(name, 0);
-	name_string = lu_value_strdup(value);
-	dn = lu_ldap_ent_to_dn(module, namingAttr, name_string, branch);
-	g_free(name_string);
+	dn = lu_ldap_ent_to_dn(module, namingAttr, name, branch);
+	g_free(name);
 	/* Process the removal. */
 #ifdef DEBUG
 	g_message("Removing `%s'.\n", dn);
@@ -1475,6 +1461,23 @@ lu_ldap_del(struct lu_module *module, enum lu_entity_type type,
 	return ret;
 }
 
+/* Return TRUE if pw starts with a valid scheme specification */
+static gboolean
+userPassword_has_scheme(const char *pw)
+{
+#define ALPHA "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
+	static const char alpha[2*26] = ALPHA;
+	static const char keystring_chars[] = ALPHA "0123456789-;";
+#undef ALPHA
+
+	/* { keystring }, keystring is defined in RFC2252. */
+	if (*pw != '{' || memchr(alpha, pw[1], sizeof(alpha)) == NULL)
+		return FALSE;
+	pw += 2;
+	pw += strspn(pw, keystring_chars);
+	return *pw == '}';
+}
+
 /* Lock an account of some kind. */
 static gboolean
 lu_ldap_handle_lock(struct lu_module *module, struct lu_ent *ent,
@@ -1484,9 +1487,7 @@ lu_ldap_handle_lock(struct lu_module *module, struct lu_ent *ent,
 	const char *dn;
 	gboolean ret = FALSE;
 	LDAPMod mod[2], *mods[3];
-	GValueArray *name, *password;
-	GValue *value;
-	char *result, *name_string, *oldpassword, *values[2][2];
+	char *result, *name, *oldpassword, *values[2][2];
 	const char *tmp, *attribute;
 	struct lu_ldap_context *ctx;
 	int err;
@@ -1499,7 +1500,7 @@ lu_ldap_handle_lock(struct lu_module *module, struct lu_ent *ent,
 	ctx = module->module_context;
 
 	/* Get the entry's name. */
-	name = lu_ent_get(ent, namingAttr);
+	name = lu_ent_get_first_value_strdup(ent, namingAttr);
 	if (name == NULL) {
 		lu_error_new(error, lu_error_generic,
 			     _("object has no %s attribute"), namingAttr);
@@ -1507,29 +1508,33 @@ lu_ldap_handle_lock(struct lu_module *module, struct lu_ent *ent,
 	}
 
 	/* Convert the name to a distinguished name. */
-	value = g_value_array_get_nth(name, 0);
-	name_string = lu_value_strdup(value);
-	dn = lu_ldap_ent_to_dn(module, namingAttr, name_string, branch);
-	g_free(name_string);
+	dn = lu_ldap_ent_to_dn(module, namingAttr, name, branch);
+	g_free(name);
 
 	attribute = ent->type == lu_user ? LU_USERPASSWORD : LU_GROUPPASSWORD;
 
-	/* Get the values for the entry's password. */
-	password = lu_ent_get_current(ent, attribute);
-	if (password == NULL) {
+	/* Get the values for the entry's password.
+	   Note that this handles only one userPassword value! */
+	oldpassword = lu_ent_get_first_value_strdup_current(ent, attribute);
+	if (oldpassword == NULL) {
 		lu_error_new(error, lu_error_generic,
 			     _("object has no %s attribute"),
 			     LU_USERPASSWORD);
 		return FALSE;
 	}
-	value = g_value_array_get_nth(password, 0);
-	oldpassword = lu_value_strdup(value);
 
 	/* We only know how to lock crypted passwords, so crypt it if it
 	 * isn't already. */
 	if (!g_str_has_prefix(oldpassword, LU_CRYPTED)) {
 		char *salt;
 
+		if (userPassword_has_scheme(oldpassword)) {
+			lu_error_new(error, lu_error_generic,
+				     _("unsupported password encryption "
+				       "scheme"));
+			g_free(oldpassword);
+			return FALSE;
+		}
 		salt = lu_util_default_salt_specifier(module->lu_context);
 		tmp = lu_make_crypted(oldpassword, salt);
 		g_free(salt);
@@ -1615,9 +1620,7 @@ lu_ldap_is_locked(struct lu_module *module, struct lu_ent *ent,
 	static const char mapped_password[] = "userPassword";
 
 	const char *dn;
-	GValueArray *name;
-	GValue *value;
-	char *name_string;
+	char *name;
 	struct lu_ldap_context *ctx = module->module_context;
 	char *attributes[] = { NULL, NULL };
 	BerValue **values;
@@ -1626,7 +1629,7 @@ lu_ldap_is_locked(struct lu_module *module, struct lu_ent *ent,
 	gboolean locked;
 
 	/* Get the name of the user or group. */
-	name = lu_ent_get(ent, namingAttr);
+	name = lu_ent_get_first_value_strdup(ent, namingAttr);
 	if (name == NULL) {
 		lu_error_new(error, lu_error_generic,
 			     _("object has no %s attribute"), namingAttr);
@@ -1634,10 +1637,8 @@ lu_ldap_is_locked(struct lu_module *module, struct lu_ent *ent,
 	}
 
 	/* Convert the name to a distinguished name. */
-	value = g_value_array_get_nth(name, 0);
-	name_string = lu_value_strdup(value);
-	dn = lu_ldap_ent_to_dn(module, namingAttr, name_string, branch);
-	g_free(name_string);
+	dn = lu_ldap_ent_to_dn(module, namingAttr, name, branch);
+	g_free(name);
 #ifdef DEBUG
 	g_print("Looking up `%s'.\n", dn);
 #endif
@@ -1678,7 +1679,7 @@ lu_ldap_is_locked(struct lu_module *module, struct lu_ent *ent,
 		val = values[i];
 		prefix_len = strlen(LU_CRYPTED);
 #ifdef DEBUG
-		g_print("Got `%s' = `.*%s'.\n", mapped_password,
+		g_print("Got `%s' = `%.*s'.\n", mapped_password,
 			(int)val->bv_len, val->bv_val);
 #endif
 		if (val->bv_len >= prefix_len
@@ -1706,9 +1707,7 @@ lu_ldap_setpass(struct lu_module *module, const char *namingAttr,
 	static const char mapped_password[] = "userPassword";
 
 	const char *dn;
-	GValueArray *name;
-	GValue *value;
-	char *name_string;
+	char *name;
 	struct lu_ldap_context *ctx = module->module_context;
 	char *attributes[] = { NULL, NULL };
 	char *addvalues[] = { NULL, NULL }, *rmvalues[] = { NULL, NULL };
@@ -1725,7 +1724,7 @@ lu_ldap_setpass(struct lu_module *module, const char *namingAttr,
 #ifdef DEBUG
 	g_print("Setting password to `%s'.\n", password);
 #endif
-	name = lu_ent_get(ent, namingAttr);
+	name = lu_ent_get_first_value_strdup(ent, namingAttr);
 	if (name == NULL) {
 		lu_error_new(error, lu_error_generic,
 			     _("object has no %s attribute"), namingAttr);
@@ -1733,16 +1732,14 @@ lu_ldap_setpass(struct lu_module *module, const char *namingAttr,
 	}
 
 	/* Convert the name to a distinguished name. */
-	value = g_value_array_get_nth(name, 0);
-	name_string = lu_value_strdup(value);
-	dn = lu_ldap_ent_to_dn(module, namingAttr, name_string, branch);
+	dn = lu_ldap_ent_to_dn(module, namingAttr, name, branch);
 #ifdef DEBUG
 	g_print("Setting password for `%s'.\n", dn);
 #endif
 
 	snprintf(filter, sizeof(filter), "(%s=%s)",
-		 map_to_ldap(module->scache, namingAttr), name_string);
-	g_free(name_string);
+		 map_to_ldap(module->scache, namingAttr), name);
+	g_free(name);
 
 	previous = NULL;
 	values = NULL;
@@ -1798,14 +1795,18 @@ lu_ldap_setpass(struct lu_module *module, const char *namingAttr,
 		const char *crypted;
 		char *salt, *tmp;
 
-		if (previous != NULL)
+		if (previous != NULL
+		    && strcmp(previous + strlen(LU_CRYPTED),
+			      LU_COMMON_DEFAULT_PASSWORD) != 0) {
 			salt = previous + strlen(LU_CRYPTED);
-		else
+			if (*salt == LOCKCHAR)
+				salt++;
+			salt = g_strdup(salt);
+		} else
 			salt = lu_util_default_salt_specifier(module
 							      ->lu_context);
 		crypted = lu_make_crypted(password, salt);
-		if (previous == NULL)
-			g_free(salt);
+		g_free(salt);
 		if (crypted == NULL) {
 			lu_error_new(error, lu_error_generic,
 				     _("error encrypting password"));
@@ -2185,6 +2186,13 @@ lu_ldap_user_default(struct lu_module *module,
 		     const char *user, gboolean is_system,
 		     struct lu_ent *ent, struct lu_error **error)
 {
+	if (lu_ent_get(ent, LU_USERPASSWORD) == NULL)
+		lu_ent_set_string(ent, LU_USERPASSWORD,
+				  LU_CRYPTED LU_COMMON_DEFAULT_PASSWORD);
+	/* This will set LU_SHADOWPASSWORD, which we ignore.  The default
+	   LU_USERPASSWORD value, which is incompatibly formated by
+	   lu_common_user_default(), will not be used because LU_USERPASSWORD
+	   was set above. */
 	return lu_common_user_default(module, user, is_system, ent, error) &&
 	       lu_common_suser_default(module, user, is_system, ent, error);
 }
@@ -2194,6 +2202,8 @@ lu_ldap_group_default(struct lu_module *module,
 		      const char *group, gboolean is_system,
 		      struct lu_ent *ent, struct lu_error **error)
 {
+	/* This sets LU_SHADOWPASSWORD, which is ignored by our backend.
+	   LU_GROUPPASSWORD is not set. */
 	return lu_common_group_default(module, group, is_system, ent, error) &&
 	       lu_common_sgroup_default(module, group, is_system, ent, error);
 }
@@ -2262,50 +2272,26 @@ lu_ldap_users_enumerate_by_group(struct lu_module *module,
 				 struct lu_error **error)
 {
 	struct lu_ldap_context *ctx;
-	GValueArray *primaries;
-	GValue *value;
+	GValueArray *ret;
 	char grp[sizeof (gid) * CHAR_BIT + 1];
-	size_t i;
 
 	LU_ERROR_CHECK(error);
 	ctx = module->module_context;
 	sprintf(grp, "%jd", (intmax_t)gid);
 
-	primaries = lu_ldap_enumerate(module, "gidNumber", grp, "uid",
-				      ctx->user_branch, error);
-	if ((error == NULL) || (*error == NULL)) {
+	ret = lu_ldap_enumerate(module, "gidNumber", grp, "uid",
+				ctx->user_branch, error);
+	if (*error == NULL) {
 		GValueArray *secondaries;
 
 		secondaries = lu_ldap_enumerate(module, "cn", group,
 						"memberUid", ctx->group_branch,
 						error);
-		for (i = 0; i < secondaries->n_values; i++) {
-			value = g_value_array_get_nth(secondaries, i);
-			g_value_array_append(primaries, value);
-		}
+		lu_util_append_values(ret, secondaries);
 		g_value_array_free(secondaries);
 	}
 
-#ifdef DEBUG
-	for (i = 0; i < primaries->n_values; i++) {
-		value = g_value_array_get_nth(primaries, i);
-		g_print("`%s' contains `%s'\n", group,
-			g_value_get_string(value));
-	}
-#endif
-	return primaries;
-}
-
-static GPtrArray *
-lu_ldap_users_enumerate_by_group_full(struct lu_module *module,
-				      const char *group, gid_t gid,
-				      struct lu_error **error)
-{
-	(void)module;
-	(void)group;
-	(void)gid;
-	(void)error;
-	return NULL;
+	return ret;
 }
 
 /* Get a list of all groups to which the user belongs, via either primary or
@@ -2317,7 +2303,7 @@ lu_ldap_groups_enumerate_by_user(struct lu_module *module,
 				 struct lu_error **error)
 {
 	struct lu_ldap_context *ctx;
-	GValueArray *primaries, *gids;
+	GValueArray *ret, *gids;
 	GValue *value;
 	size_t i;
 
@@ -2326,7 +2312,7 @@ lu_ldap_groups_enumerate_by_user(struct lu_module *module,
 	ctx = module->module_context;
 
 	/* Create an array to hold the values returned. */
-	primaries = g_value_array_new(0);
+	ret = g_value_array_new(0);
 
 	/* Get the user's primary GID(s). */
 	gids = lu_ldap_enumerate(module, "uid", user, "gidNumber",
@@ -2338,60 +2324,39 @@ lu_ldap_groups_enumerate_by_user(struct lu_module *module,
 
 		value = g_value_array_get_nth(gids, i);
 		gid = lu_value_get_id(value);
-		g_assert (gid != LU_VALUE_INVALID_ID);
+		if (gid == LU_VALUE_INVALID_ID)
+			continue;
 		ent = lu_ent_new();
 		if (lu_group_lookup_id(module->lu_context, gid,
-				       ent, error)) {
-			GValueArray *values;
-			size_t j;
-
+				       ent, error))
 			/* Get the group's names and add them to the list
 			 * of values to return. */
-			values = lu_ent_get(ent, LU_GROUPNAME);
-			for (j = 0; j < values->n_values; j++) {
-				value = g_value_array_get_nth(values, j);
-				g_value_array_append(primaries, value);
-			}
-		}
+			lu_util_append_values(ret,
+					      lu_ent_get(ent, LU_GROUPNAME));
 		lu_ent_free(ent);
 	}
 	g_value_array_free(gids);
 	/* Search for the supplemental groups which list this user as
 	 * a member. */
-	if ((error == NULL) || (*error == NULL)) {
+	if (*error == NULL) {
 		GValueArray *secondaries;
 
 		secondaries = lu_ldap_enumerate(module, "memberUid", user,
 						"cn", ctx->group_branch,
 						error);
-		for (i = 0; i < secondaries->n_values; i++) {
-			value = g_value_array_get_nth(secondaries, i);
-			g_value_array_append(primaries, value);
-		}
+		lu_util_append_values(ret, secondaries);
 		g_value_array_free(secondaries);
 	}
 
 #ifdef DEBUG
-	for (i = 0; i < primaries->n_values; i++) {
-		value = g_value_array_get_nth(primaries, i);
+	for (i = 0; i < ret->n_values; i++) {
+		value = g_value_array_get_nth(ret, i);
 		g_print("`%s' is in `%s'\n", user,
 			g_value_get_string(value));
 	}
 #endif
 
-	return primaries;
-}
-
-static GPtrArray *
-lu_ldap_groups_enumerate_by_user_full(struct lu_module *module,
-				      const char *user, uid_t uid,
-				      struct lu_error **error)
-{
-	(void)module;
-	(void)user;
-	(void)uid;
-	(void)error;
-	return NULL;
+	return ret;
 }
 
 static gboolean
@@ -2422,6 +2387,34 @@ lu_ldap_close_module(struct lu_module *module)
 	return TRUE;
 }
 
+static gboolean
+lu_ldap_valid_module_combination(struct lu_module *module, GValueArray *names,
+				 struct lu_error **error)
+{
+	size_t i;
+
+	g_assert(module != NULL);
+	g_assert(names != NULL);
+	LU_ERROR_CHECK(error);
+	for (i = 0; i < names->n_values; i++) {
+		const char *name;
+
+		name = g_value_get_string(g_value_array_get_nth(names, i));
+		if (strcmp(name, LU_MODULE_NAME_FILES) == 0
+		    || strcmp(name, LU_MODULE_NAME_SHADOW) == 0) {
+			/* These modules use an incompatible LU_*PASSWORD
+			   format: the LU_CRYPTED prefix, or a similar
+			   indicator of an LDAP-defined hashing method, is
+			   missing. */
+			lu_error_new(error, lu_error_invalid_module_combination,
+				     _("the `%s' and `%s' modules can not be "
+				       "combined"), module->name, name);
+			return FALSE;
+		}
+	}
+	return TRUE;
+}
+
 static gboolean
 lu_ldap_uses_elevated_privileges(struct lu_module *module)
 {
@@ -2470,6 +2463,8 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 
 	ctx->prompts[LU_LDAP_PASSWORD].key = "ldap/password";
 	ctx->prompts[LU_LDAP_PASSWORD].prompt = N_("LDAP Bind Password");
+	ctx->prompts[LU_LDAP_PASSWORD].default_value =
+		lu_cfg_read_single(context, "ldap/password", NULL);
 	ctx->prompts[LU_LDAP_PASSWORD].visible = FALSE;
 
 	ctx->prompts[LU_LDAP_AUTHUSER].key = "ldap/user";
@@ -2539,7 +2534,7 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	ret->version = LU_MODULE_VERSION;
 	ret->module_context = ctx;
 	ret->scache = lu_string_cache_new(TRUE);
-	ret->name = ret->scache->cache(ret->scache, "ldap");
+	ret->name = ret->scache->cache(ret->scache, LU_MODULE_NAME_LDAP);
 	ctx->module = ret;
 
 	ctx->user_branch = lu_cfg_read_single(context, "ldap/userBranch",
@@ -2558,8 +2553,8 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	ctx->ldap = ldap;
 
 	ctx->mapped_user_attributes
-		= g_malloc0(sizeof(*ctx->mapped_user_attributes)
-			    * G_N_ELEMENTS(lu_ldap_user_attributes));
+		= g_malloc0_n(G_N_ELEMENTS(lu_ldap_user_attributes),
+			      sizeof(*ctx->mapped_user_attributes));
 	for (i = 0; i < G_N_ELEMENTS(lu_ldap_user_attributes); i++) {
 		if (lu_ldap_user_attributes[i] != NULL)
 			ctx->mapped_user_attributes[i] = (char *)
@@ -2570,8 +2565,8 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	}
 
 	ctx->mapped_group_attributes
-		= g_malloc0(sizeof(*ctx->mapped_group_attributes)
-			    * G_N_ELEMENTS(lu_ldap_group_attributes));
+		= g_malloc0_n(G_N_ELEMENTS(lu_ldap_group_attributes),
+			      sizeof(*ctx->mapped_group_attributes));
 	for (i = 0; i < G_N_ELEMENTS(lu_ldap_group_attributes); i++) {
 		if (lu_ldap_group_attributes[i] != NULL)
 			ctx->mapped_group_attributes[i] = (char *)
@@ -2582,6 +2577,7 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	}
 
 	/* Set the method pointers. */
+	ret->valid_module_combination = lu_ldap_valid_module_combination;
 	ret->uses_elevated_privileges = lu_ldap_uses_elevated_privileges;
 
 	ret->user_lookup_name = lu_ldap_user_lookup_name;
@@ -2600,7 +2596,6 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	ret->users_enumerate = lu_ldap_users_enumerate;
 	ret->users_enumerate_by_group = lu_ldap_users_enumerate_by_group;
 	ret->users_enumerate_full = lu_ldap_users_enumerate_full;
-	ret->users_enumerate_by_group_full = lu_ldap_users_enumerate_by_group_full;
 
 	ret->group_lookup_name = lu_ldap_group_lookup_name;
 	ret->group_lookup_id = lu_ldap_group_lookup_id;
@@ -2618,7 +2613,6 @@ libuser_ldap_init(struct lu_context *context, struct lu_error **error)
 	ret->groups_enumerate = lu_ldap_groups_enumerate;
 	ret->groups_enumerate_by_user = lu_ldap_groups_enumerate_by_user;
 	ret->groups_enumerate_full = lu_ldap_groups_enumerate_full;
-	ret->groups_enumerate_by_user_full = lu_ldap_groups_enumerate_by_user_full;
 
 	ret->close = lu_ldap_close_module;
 
