@@ -12,12 +12,14 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <Python.h>
 #include <config.h>
+#include <errno.h>
 #include <grp.h>
+#include <inttypes.h>
 #include <pwd.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -400,6 +402,7 @@ libuser_admin_wrap_boolean(PyObject *self, PyObject *args, PyObject *kwargs,
 	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!", keywords,
 					 &EntityType, &ent)) {
 		DEBUG_EXIT;
+		return NULL;
 	}
 	/* Run the function. */
 	ret = fn(me->ctx, ent->ent, &error);
@@ -457,8 +460,6 @@ libuser_admin_create_home(PyObject *self, PyObject *args,
 	struct libuser_entity *ent = NULL;
 	struct lu_context *context;
 	const char *dir, *skeleton = NULL;
-	GValueArray *values;
-	GValue *value;
 	char *keywords[] = { "home", "skeleton", NULL };
 	uid_t uidNumber = 0;
 	gid_t gidNumber = 0;
@@ -477,37 +478,31 @@ libuser_admin_create_home(PyObject *self, PyObject *args,
 	}
 
 	/* Get the user's home directory value. */
-	values = lu_ent_get(ent->ent, LU_HOMEDIRECTORY);
-	if (values == NULL) {
+	dir = lu_ent_get_first_string(ent->ent, LU_HOMEDIRECTORY);
+	if (dir == NULL) {
 		PyErr_SetString(PyExc_KeyError,
 				"user does not have a `" LU_HOMEDIRECTORY
 				"' attribute");
 		return NULL;
 	}
-	value = g_value_array_get_nth(values, 0);
-	dir = g_value_get_string(value);
 
 	/* Get the user's UID. */
-	values = lu_ent_get(ent->ent, LU_UIDNUMBER);
-	if (values == NULL) {
+	uidNumber = lu_ent_get_first_id(ent->ent, LU_UIDNUMBER);
+	if (uidNumber == LU_VALUE_INVALID_ID) {
 		PyErr_SetString(PyExc_KeyError,
 				"user does not have a `" LU_UIDNUMBER
 				"' attribute");
 		return NULL;
 	}
-	value = g_value_array_get_nth(values, 0);
-	uidNumber = lu_value_get_id(value);
 
 	/* Get the user's GID. */
-	values = lu_ent_get(ent->ent, LU_GIDNUMBER);
-	if (values == NULL) {
+	gidNumber = lu_ent_get_first_id(ent->ent, LU_GIDNUMBER);
+	if (gidNumber == LU_VALUE_INVALID_ID) {
 		PyErr_SetString(PyExc_KeyError,
 				"user does not have a `" LU_GIDNUMBER
 				"' attribute");
 		return NULL;
 	}
-	value = g_value_array_get_nth(values, 0);
-	gidNumber = lu_value_get_id(value);
 
 	/* Attempt to populate the directory. */
 	if (lu_homedir_populate(context, skeleton, dir, uidNumber, gidNumber,
@@ -535,9 +530,6 @@ libuser_admin_remove_home(PyObject *self, PyObject *args,
 			  PyObject *kwargs)
 {
 	struct libuser_entity *ent = NULL;
-	const char *dir;
-	GValueArray *values;
-	GValue *value;
 	char *keywords[] = { "home", NULL };
 	struct lu_error *error = NULL;
 
@@ -551,19 +543,47 @@ libuser_admin_remove_home(PyObject *self, PyObject *args,
 		return NULL;
 	}
 
-	/* Get the user's home directory. */
-	values = lu_ent_get(ent->ent, LU_HOMEDIRECTORY);
-	if (values == NULL) {
-		PyErr_SetString(PyExc_KeyError,
-				"user does not have a `" LU_HOMEDIRECTORY
-				"' attribute");
+	/* Remove the directory. */
+	if (lu_homedir_remove_for_user(ent->ent, &error)) {
+		/* Successfully removed. */
+		DEBUG_EXIT;
+		return PyInt_FromLong(1);
+	} else {
+		/* Removal failed.  You'll have to come back for repeated
+		 * treatments. */
+		PyErr_SetString(PyExc_RuntimeError,
+				error ?
+				error->string :
+				_("error removing home directory for user"));
+		if (error) {
+			lu_error_free(&error);
+		}
+		DEBUG_EXIT;
+		return NULL;
+	}
+}
+
+/* Remove a user's home directory if it is owned by them. */
+static PyObject *
+libuser_admin_remove_home_if_owned(PyObject *self, PyObject *args,
+				   PyObject *kwargs)
+{
+	struct libuser_entity *ent = NULL;
+	char *keywords[] = { "user", NULL };
+	struct lu_error *error = NULL;
+
+	(void)self;
+	DEBUG_ENTRY;
+
+	/* We expect an object. */
+	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O!", keywords,
+					 &EntityType, &ent)) {
+		DEBUG_EXIT;
 		return NULL;
 	}
-	value = g_value_array_get_nth(values, 0);
-	dir = g_value_get_string(value);
 
 	/* Remove the directory. */
-	if (lu_homedir_remove(dir, &error)) {
+	if (lu_homedir_remove_for_user_if_owned(ent->ent, &error)) {
 		/* Successfully removed. */
 		DEBUG_EXIT;
 		return PyInt_FromLong(1);
@@ -589,8 +609,6 @@ libuser_admin_move_home(PyObject *self, PyObject *args,
 {
 	struct libuser_entity *ent = NULL;
 	const char *olddir = NULL, *newdir = NULL;
-	GValueArray *values;
-	GValue *value;
 	char *keywords[] = { "entity", "newhome", NULL };
 	struct lu_error *error = NULL;
 
@@ -607,38 +625,33 @@ libuser_admin_move_home(PyObject *self, PyObject *args,
 	if (newdir != NULL) {
 		/* We were given a string, so move the user's home directory
 		 * to the new location. */
-		values = lu_ent_get(ent->ent, LU_HOMEDIRECTORY);
-		if (values == NULL) {
+		olddir = lu_ent_get_first_string(ent->ent, LU_HOMEDIRECTORY);
+		if (olddir == NULL) {
 			PyErr_SetString(PyExc_KeyError,
 					"user does not have a current `"
 					LU_HOMEDIRECTORY "' attribute");
 			return NULL;
 		}
-		value = g_value_array_get_nth(values, 0);
-		olddir = g_value_get_string(value);
 	} else {
 		/* We weren't given a string, so use the current and pending
 		 * values, and move from one to the other. */
-		values = lu_ent_get_current(ent->ent, LU_HOMEDIRECTORY);
-		if (values == NULL) {
+		olddir = lu_ent_get_first_string_current(ent->ent,
+							 LU_HOMEDIRECTORY);
+		if (olddir == NULL) {
 			PyErr_SetString(PyExc_KeyError,
 					"user does not have a current `"
 					LU_HOMEDIRECTORY "' attribute");
 			return NULL;
 		}
-		value = g_value_array_get_nth(values, 0);
-		olddir = g_value_get_string(value);
 
 		/* Now read the pending directory. */
-		values = lu_ent_get(ent->ent, LU_HOMEDIRECTORY);
-		if (values == NULL) {
+		newdir = lu_ent_get_first_string(ent->ent, LU_HOMEDIRECTORY);
+		if (newdir == NULL) {
 			PyErr_SetString(PyExc_KeyError,
 					"user does not have a pending `"
 					LU_HOMEDIRECTORY "' attribute");
 			return NULL;
 		}
-		value = g_value_array_get_nth(values, 0);
-		newdir = g_value_get_string(value);
 	}
 
 	/* Attempt the move. */
@@ -669,6 +682,8 @@ libuser_admin_create_remove_mail(PyObject *self, PyObject *args,
 
 	char *keywords[] = { "entity", NULL };
 	struct libuser_admin *me = (struct libuser_admin *) self;
+	struct lu_error *error;
+	gboolean res;
 
 	DEBUG_ENTRY;
 
@@ -680,11 +695,17 @@ libuser_admin_create_remove_mail(PyObject *self, PyObject *args,
 	}
 
 	/* Now just pass it to the internal function. */
-	if (lu_mailspool_create_remove(me->ctx, ent->ent, action)) {
+	error = NULL;
+	if (action)
+		res = lu_mail_spool_create(me->ctx, ent->ent, &error);
+	else
+		res = lu_mail_spool_remove(me->ctx, ent->ent, &error);
+	if (res) {
 		return PyInt_FromLong(1);
 	} else {
-		PyErr_SetString(PyExc_RuntimeError,
-				_("error creating mail spool for user"));
+		PyErr_SetString(PyExc_RuntimeError, lu_strerror(error));
+		if (error != NULL)
+			lu_error_free(&error);
 		return NULL;
 	}
 }
@@ -756,13 +777,16 @@ libuser_admin_add_user(PyObject *self, PyObject *args, PyObject *kwargs)
 	}
 	if (ret != NULL && mkmailspool != NULL
 	    && PyObject_IsTrue(mkmailspool)) {
+		struct lu_error *error;
+
 		Py_DECREF(ret);
-		if (lu_mailspool_create_remove(context, ent->ent, TRUE))
+		error = NULL;
+		if (lu_mail_spool_create(context, ent->ent, &error))
 			ret = PyInt_FromLong(1);
 		else {
-			PyErr_SetString(PyExc_RuntimeError,
-					_("error creating mail spool for "
-					  "user"));
+			PyErr_SetString(PyExc_RuntimeError, lu_strerror(error));
+			if (error != NULL)
+				lu_error_free(&error);
 			ret = NULL;
 		}
 	}
@@ -875,15 +899,17 @@ libuser_admin_delete_user(PyObject *self, PyObject *args,
 	if (ret != NULL && rmmailspool != NULL
 	    && PyObject_IsTrue(rmmailspool)) {
 		struct libuser_entity *entity;
+		struct lu_error *error;
 
 		Py_DECREF(ret);
 		entity = (struct libuser_entity *)ent;
-		if (lu_mailspool_create_remove(context, entity->ent, TRUE))
+		error = NULL;
+		if (lu_mail_spool_remove(context, entity->ent, &error))
 			ret = PyInt_FromLong(1);
 		else {
-			PyErr_SetString(PyExc_RuntimeError,
-					_("error removing mail spool for "
-					  "user"));
+			PyErr_SetString(PyExc_RuntimeError, lu_strerror(error));
+			if (error != NULL)
+				lu_error_free(&error);
 			ret = NULL;
 		}
 	}
@@ -1155,7 +1181,6 @@ libuser_admin_enumerate_users_full(PyObject *self, PyObject *args,
 	struct lu_error *error = NULL;
 	char *keywords[] = { "pattern", NULL };
 	struct libuser_admin *me = (struct libuser_admin *) self;
-	size_t i;
 
 	DEBUG_ENTRY;
 	/* Expect a possible pattern. */
@@ -1169,15 +1194,9 @@ libuser_admin_enumerate_users_full(PyObject *self, PyObject *args,
 	if (error != NULL)
 		lu_error_free(&error);
 	/* Convert the list to a PyList. */
-	ret = PyList_New(0);
-	for (i = 0; i < results->len; i++) {
-		PyObject *ent;
-
-		ent = libuser_wrap_ent(g_ptr_array_index(results, i));
-		PyList_Append(ret, ent);
-		Py_DECREF(ent);
-	}
-	g_ptr_array_free(results, TRUE);
+	ret = convert_ent_array_pylist(results);
+	if (results != NULL)
+		g_ptr_array_free(results, TRUE);
 	DEBUG_EXIT;
 	return ret;
 }
@@ -1193,7 +1212,6 @@ libuser_admin_enumerate_groups_full(PyObject *self, PyObject *args,
 	struct lu_error *error = NULL;
 	char *keywords[] = { "pattern", NULL };
 	struct libuser_admin *me = (struct libuser_admin *) self;
-	size_t i;
 
 	DEBUG_ENTRY;
 	/* Possibly expect a pattern. */
@@ -1207,20 +1225,13 @@ libuser_admin_enumerate_groups_full(PyObject *self, PyObject *args,
 	if (error != NULL)
 		lu_error_free(&error);
 	/* Convert the list to a PyList. */
-	ret = PyList_New(0);
-	for (i = 0; i < results->len; i++) {
-		PyObject *ent;
-
-		ent = libuser_wrap_ent(g_ptr_array_index(results, i));
-		PyList_Append(ret, ent);
-		Py_DECREF(ent);
-	}
-	g_ptr_array_free(results, TRUE);
+	ret = convert_ent_array_pylist(results);
+	if (results != NULL)
+		g_ptr_array_free(results, TRUE);
 	DEBUG_EXIT;
 	return ret;
 }
 
-#ifdef PLACEHOLDERS
 /* Get the list of users who belong to a group. */
 static PyObject *
 libuser_admin_enumerate_users_by_group_full(PyObject *self, PyObject *args,
@@ -1229,8 +1240,9 @@ libuser_admin_enumerate_users_by_group_full(PyObject *self, PyObject *args,
 	GPtrArray *results;
 	char *group = NULL;
 	PyObject *ret;
+	struct lu_error *error = NULL;
 	char *keywords[] = { "group", NULL };
-	int i;
+	struct libuser_admin *me = (struct libuser_admin *) self;
 
 	DEBUG_ENTRY;
 	/* Expect the group's name. */
@@ -1239,15 +1251,12 @@ libuser_admin_enumerate_users_by_group_full(PyObject *self, PyObject *args,
 		return NULL;
 	}
 	/* Get a list of the users in this group. */
-	ret = PyList_New(0);
-	for (i = 0; i < results->len; i++) {
-		PyObject *ent;
-
-		ent = libuser_wrap_ent(g_ptr_array_index(results, i));
-		PyList_Append(ret, ent);
-		Py_DECREF(ent);
-	}
-	g_ptr_array_free(results, TRUE);
+	results = lu_users_enumerate_by_group_full(me->ctx, group, &error);
+	if (error != NULL)
+		lu_error_free(&error);
+	ret = convert_ent_array_pylist(results);
+	if (results != NULL)
+		g_ptr_array_free(results, TRUE);
 	DEBUG_EXIT;
 	return ret;
 }
@@ -1263,7 +1272,6 @@ libuser_admin_enumerate_groups_by_user_full(PyObject *self, PyObject *args,
 	struct lu_error *error = NULL;
 	char *keywords[] = { "user", NULL };
 	struct libuser_admin *me = (struct libuser_admin *) self;
-	int i;
 
 	DEBUG_ENTRY;
 	/* Expect the user's name. */
@@ -1272,33 +1280,54 @@ libuser_admin_enumerate_groups_by_user_full(PyObject *self, PyObject *args,
 		return NULL;
 	}
 	/* Get the list. */
-	ret = PyList_New(0);
-	for (i = 0; i < results->len; i++) {
-		PyObject *ent;
-
-		ent = libuser_wrap_ent(g_ptr_array_index(results, i));
-		PyList_Append(ret, ent);
-		Py_DECREF(ent);
-	}
-	g_ptr_array_free(results, TRUE);
 	results = lu_groups_enumerate_by_user_full(me->ctx, user, &error);
+	if (error != NULL)
+		lu_error_free(&error);
+	ret = convert_ent_array_pylist(results);
+	if (results != NULL)
+		g_ptr_array_free(results, TRUE);
 	DEBUG_EXIT;
 	return ret;
 }
-#endif
 
 static PyObject *
 libuser_admin_get_first_unused_id_type(struct libuser_admin *me,
 				       PyObject * args, PyObject * kwargs,
 				       enum lu_entity_type enttype)
 {
+	const char *key, *key_string, *val;
 	char *keywords[] = { "start", NULL };
-
 	PY_LONG_LONG start = 500;
 
 	g_return_val_if_fail(me != NULL, NULL);
 
 	DEBUG_ENTRY;
+
+	switch (enttype) {
+	case lu_user:
+		key = "userdefaults/" LU_UIDNUMBER;
+		key_string = "userdefaults/" G_STRINGIFY_ARG(LU_UIDNUMBER);
+		break;
+	case lu_group:
+		key = "groupdefaults/" LU_GIDNUMBER;
+		key_string = "groupdefaults/" G_STRINGIFY_ARG(LU_GIDNUMBER);
+		break;
+	default:
+		g_assert_not_reached();
+	}
+	val = lu_cfg_read_single(me->ctx, key, NULL);
+	if (val == NULL)
+		val = lu_cfg_read_single(me->ctx, key_string, NULL);
+	if (val != NULL) {
+		intmax_t imax;
+		char *end;
+
+		errno = 0;
+		imax = strtoimax(val, &end, 10);
+		if (errno == 0 && *end == 0 && end != val && (id_t)imax == imax)
+			start = imax;
+	}
+
 	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|L", keywords,
 					 &start)) {
 		DEBUG_EXIT;
@@ -1427,7 +1456,6 @@ static struct PyMethodDef libuser_admin_methods[] = {
 	{"enumerateGroupsFull", (PyCFunction) libuser_admin_enumerate_groups_full,
 	 METH_VARARGS | METH_KEYWORDS,
 	 "get a list of groups matching a pattern, in listed databases"},
-#ifdef PLACEHOLDERS
 	{"enumerateUsersByGroupFull",
 	 (PyCFunction) libuser_admin_enumerate_users_by_group_full,
 	 METH_VARARGS | METH_KEYWORDS,
@@ -1436,7 +1464,6 @@ static struct PyMethodDef libuser_admin_methods[] = {
 	 (PyCFunction) libuser_admin_enumerate_groups_by_user_full,
 	 METH_VARARGS | METH_KEYWORDS,
 	 "get a list of groups to which a user belongs"},
-#endif
 
 	{"promptConsole", (PyCFunction) libuser_admin_prompt_console,
 	 METH_VARARGS | METH_KEYWORDS,
@@ -1455,6 +1482,9 @@ static struct PyMethodDef libuser_admin_methods[] = {
 	{"removeHome", (PyCFunction) libuser_admin_remove_home,
 	 METH_VARARGS | METH_KEYWORDS,
 	 "remove a user's home directory"},
+	{"removeHomeIfOwned", (PyCFunction) libuser_admin_remove_home_if_owned,
+	 METH_VARARGS | METH_KEYWORDS,
+	 "remove a user's home directory if it is owned by them"},
 
 	{"createMail", (PyCFunction) libuser_admin_create_mail,
 	 METH_VARARGS | METH_KEYWORDS,
