@@ -12,7 +12,7 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <Python.h>
@@ -91,6 +91,29 @@ convert_value_array_pylist(GValueArray *array)
 	return ret;
 }
 
+/* Convert a (potentially NULL) GPtrArray of entities into a Python list of
+   values. */
+static PyObject *
+convert_ent_array_pylist(GPtrArray *array)
+{
+	PyObject *ret;
+	size_t i;
+
+	DEBUG_ENTRY;
+
+	ret = PyList_New(0);
+	for (i = 0; array != NULL && i < array->len; i++) {
+		PyObject *ent;
+
+		ent = libuser_wrap_ent(g_ptr_array_index(array, i));
+		PyList_Append(ret, ent);
+		Py_DECREF(ent);
+	}
+
+	DEBUG_EXIT;
+	return ret;
+}
+
 /* Wrap up an entity object in a pretty Python wrapper. */
 static PyObject *
 libuser_wrap_ent(struct lu_ent *ent)
@@ -163,7 +186,7 @@ libuser_convert_to_value(PyObject *item, GValue *value)
 		if ((long)ll == ll) {
 			g_value_init(value, G_TYPE_LONG);
 			g_value_set_long(value, ll);
-		} else if ((id_t)ll == ll)
+		} else if ((id_t)ll == ll && (id_t)ll != LU_VALUE_INVALID_ID)
 			lu_value_init_set_id(value, ll);
 		else {
 			PyErr_SetString(PyExc_OverflowError,
@@ -214,7 +237,7 @@ libuser_convert_to_value(PyObject *item, GValue *value)
 		if ((long)ll == ll) {
 			g_value_init(value, G_TYPE_LONG);
 			g_value_set_long(value, ll);
-		} else if ((id_t)ll == ll)
+		} else if ((id_t)ll == ll && (id_t)ll != LU_VALUE_INVALID_ID)
 			lu_value_init_set_id(value, ll);
 		else {
 			PyErr_SetString(PyExc_OverflowError,
@@ -347,22 +370,22 @@ static PyObject *
 libuser_entity_getattrlist(PyObject *self, PyObject *ignore)
 {
 	struct libuser_entity *me;
-	GList *i;
+	GList *list, *i;
 	PyObject *ret;
 
 	(void)ignore;
 	DEBUG_ENTRY;
 	me = (struct libuser_entity *)self;
 	ret = PyList_New(0);
-	for (i = lu_ent_get_attributes(me->ent);
-	     i != NULL;
-	     i = g_list_next(i)) {
+	list = lu_ent_get_attributes(me->ent);
+	for (i = list; i != NULL; i = g_list_next(i)) {
 		PyObject *str;
 
 		str = PyString_FromString((char*)i->data);
 		PyList_Append(ret, str);
 		Py_DECREF(str);
 	}
+	g_list_free(list);
 	DEBUG_EXIT;
 	return ret;
 }
@@ -565,10 +588,15 @@ static Py_ssize_t
 libuser_entity_length(PyObject *self)
 {
 	struct libuser_entity *me;
+	GList *list;
+	Py_ssize_t ret;
 
 	DEBUG_CALL;
 	me = (struct libuser_entity *)self;
-	return g_list_length(lu_ent_get_attributes(me->ent));
+	list = lu_ent_get_attributes(me->ent);
+	ret = g_list_length(list);
+	g_list_free(list);
+	return ret;
 }
 
 /* Get the value for a particular item, dictionary style. */
