@@ -12,7 +12,7 @@
  *
  * You should have received a copy of the GNU Library General Public
  * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <Python.h>
@@ -55,6 +55,11 @@ libuser_admin_python_prompter(struct lu_prompt *prompts, int count,
 
 			prompt = (struct libuser_prompt *)
 			  libuser_prompt_new(NULL, NULL);
+			if (prompt == NULL) {
+				Py_DECREF(list);
+				DEBUG_EXIT;
+				return FALSE;
+			}
 			prompt->prompt.key = g_strdup(prompts[i].key);
 			prompt->prompt.prompt = g_strdup(prompts[i].prompt);
 			prompt->prompt.domain = g_strdup(prompts[i].domain);
@@ -148,7 +153,7 @@ libuser_admin_prompt(struct libuser_admin *self, PyObject * args,
 		DEBUG_CALL;
 	}
 	DEBUG_CALL;
-	prompts = g_malloc0(count * sizeof(struct lu_prompt));
+	prompts = g_malloc0_n(count, sizeof(struct lu_prompt));
 	DEBUG_CALL;
 
 	for (i = 0; i < count; i++) {
