@@ -15,13 +15,13 @@
 #
 # You should have received a copy of the GNU Library General Public
 # License along with this program; if not, write to the Free Software
-# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 #
 # Author: Miloslav Trmac <mitr@redhat.com>
 
 srcdir=$srcdir/tests
 
-workdir=$(pwd)/test_files
+workdir=$(pwd)/test_config
 
 trap 'status=$?; rm -rf "$workdir"; exit $status' 0
 trap '(exit 1); exit 1' 1 2 13 15
@@ -42,5 +42,11 @@ sed -e "s|@TOP_BUILDDIR@|$(pwd)|g" -e "s|@SRCDIR@|$srcdir|g" \
 # Ugly non-portable hacks
 LD_LIBRARY_PATH=$(pwd)/lib/.libs
 export LD_LIBRARY_PATH
+PYTHONPATH=$(pwd)/python/.libs
+export PYTHONPATH
 
-tests/config_test "$workdir"
+$VALGRIND tests/config_test "$workdir"
+
+LIBUSER_CONF="$workdir/libuser_import.conf"
+export LIBUSER_CONF
+workdir="$workdir" $VALGRIND python "$srcdir"/config_test.py
