@@ -40,10 +40,10 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.LOGINSHELL], ['/bin/bash'])
         self.assertEqual(e[libuser.SHADOWNAME], ['empty_user'])
         self.assertEqual(e[libuser.SHADOWPASSWORD], [''])
-        self.assertEqual(e[libuser.SHADOWLASTCHANGE], [1000])
-        self.assertEqual(e[libuser.SHADOWMIN], [0])
-        self.assertEqual(e[libuser.SHADOWMAX], [99999])
-        self.assertEqual(e[libuser.SHADOWWARNING], [7])
+        self.assertEqual(e[libuser.SHADOWLASTCHANGE], [-1])
+        self.assertEqual(e[libuser.SHADOWMIN], [-1])
+        self.assertEqual(e[libuser.SHADOWMAX], [-1])
+        self.assertEqual(e[libuser.SHADOWWARNING], [-1])
         self.assertEqual(e[libuser.SHADOWINACTIVE], [-1])
         self.assertEqual(e[libuser.SHADOWEXPIRE], [-1])
         self.assertEqual(e[libuser.SHADOWFLAG], [-1])
@@ -212,6 +212,56 @@ class Tests(unittest.TestCase):
         e[libuser.HOMEDIRECTORY] = 'this_is_a_relative_path'
         self.assertRaises(RuntimeError, self.a.addUser, e, mkmailspool = False)
 
+    def testUserAdd6(self):
+        # ':' in field values
+        # libuser.USERPASSWORD not tested because lu_shadow_user_add_prep()
+        # always replaces it by 'x'.  libuser.UIDNUMBER not tested because
+        # ent_has_name_and_id() interprets the value as a number.
+        for field in (libuser.USERNAME, libuser.GIDNUMBER, libuser.GECOS,
+                      libuser.HOMEDIRECTORY, libuser.SHADOWPASSWORD,
+                      libuser.SHADOWLASTCHANGE, libuser.SHADOWMIN,
+                      libuser.SHADOWMAX, libuser.SHADOWWARNING,
+                      libuser.SHADOWINACTIVE, libuser.SHADOWEXPIRE):
+            e = self.a.initUser('user_6_6' + field)
+            e[field] = str(e[field][0]) + ':x'
+            self.assertRaises(RuntimeError, self.a.addUser, e, False, False)
+        # libuser.SHADOWFLAG not tested because it is parsed as an integer
+        e = self.a.initUser('user_6_6' + libuser.LOGINSHELL)
+        e[libuser.LOGINSHELL] = e[libuser.LOGINSHELL][0] + ':x'
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user_6_6' + libuser.LOGINSHELL)
+        self.assert_(':' in e[libuser.LOGINSHELL][0])
+
+    def testUserAdd7(self):
+        # Let the user specify dangerous/dubious home directory paths explicitly
+        for name in ('.', '..'):
+            e = self.a.initUser(name)
+            e[libuser.HOMEDIRECTORY] = '/home/' + name
+            self.a.addUser(e, False, False)
+            del e
+            e = self.a.lookupUserByName(name)
+            self.assertEqual(e[libuser.HOMEDIRECTORY], ['/home/' + name])
+            self.a.deleteUser(e, False, False)
+        e = self.a.initUser('user6_7')
+        e[libuser.HOMEDIRECTORY] = '/home/..'
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_7')
+        self.assertEqual(e[libuser.HOMEDIRECTORY], ['/home/..'])
+        # ... but never create such home directory paths by default
+        for name in ('.', '..'):
+            e = self.a.initUser(name)
+            self.assertRaises(RuntimeError, self.a.addUser, e, False, False)
+
+    def testUserAdd8(self):
+        # Adding duplicate entries
+        e = self.a.initUser('user6_8')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.initUser('user6_8')
+        self.assertRaises(RuntimeError, self.a.addUser, e, False, False)
+
     def testUserMod1(self):
         # A minimal case
         e = self.a.initUser('user7_1')
@@ -313,9 +363,64 @@ class Tests(unittest.TestCase):
         del e
         e = self.a.lookupUserByName('user7_4')
         self.assert_(e)
-        self.assertEqual(e[libuser.SHADOWMAX], [99999])
+        self.assertEqual(e[libuser.SHADOWMAX], [-1])
         self.assertEqual(e[libuser.SHADOWFLAG], [-1])
 
+    def testUserMod5(self):
+        # ':' in field values
+        # libuser.USERPASSWORD not tested because lu_shadow_user_add_prep()
+        # always replaces it by 'x'.  libuser.UIDNUMBER not tested because
+        # ent_has_name_and_id() interprets the value as a number.
+        for field in (libuser.USERNAME, libuser.USERPASSWORD, libuser.GIDNUMBER,
+                      libuser.GECOS, libuser.HOMEDIRECTORY,
+                      libuser.SHADOWPASSWORD, libuser.SHADOWLASTCHANGE,
+                      libuser.SHADOWMIN, libuser.SHADOWMAX,
+                      libuser.SHADOWWARNING, libuser.SHADOWINACTIVE,
+                      libuser.SHADOWEXPIRE):
+            e = self.a.initUser('user7_5' + field)
+            self.a.addUser(e, False, False)
+            del e
+            e = self.a.lookupUserByName('user7_5' + field)
+            self.assert_(e)
+            self.assert_(':' not in str(e[field][0]))
+            e[field] = str(e[field][0]) + ':x'
+            self.assertRaises(RuntimeError, self.a.modifyUser, e, False)
+        # libuser.SHADOWFLAG not tested because it is parsed as an integer
+        e = self.a.initUser('user7_5')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user7_5')
+        self.assert_(e)
+        self.assert_(':' not in e[libuser.LOGINSHELL][0])
+        e[libuser.LOGINSHELL] = e[libuser.LOGINSHELL][0] + ':x'
+        self.a.modifyUser(e, False)
+        del e
+        e = self.a.lookupUserByName('user7_5')
+        self.assert_(e)
+        self.assert_(':' in e[libuser.LOGINSHELL][0])
+
+    def testUserMod6(self):
+        # Nothing to do with the files module: test lu_name_allowed is called.
+        e = self.a.initUser('user7_6')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user7_6')
+        e[libuser.USERNAME] = 'very_long_name_123456789_123456789_123456789'
+        self.assertRaises(RuntimeError, self.a.modifyUser, e, False)
+        del e
+
+    def testUserMod7(self):
+        # Renaming to create duplicate entries
+        e = self.a.initUser('user7_7')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.initUser('user7_8')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user7_8')
+        e[libuser.USERNAME] = 'user7_7'
+        self.assertRaises(RuntimeError, self.a.modifyUser, e, False)
+
     def testUserDel(self):
         e = self.a.initUser('user8_1')
         self.a.addUser(e, False, False)
@@ -459,6 +564,7 @@ class Tests(unittest.TestCase):
         del e
         e = self.a.lookupUserByName('user12_1')
         self.assertEqual(e[libuser.USERPASSWORD], ['x'])
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
         crypted = crypt.crypt('password', e[libuser.SHADOWPASSWORD][0])
         self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
         self.assert_(e[libuser.SHADOWLASTCHANGE][0] > 10000)
@@ -482,6 +588,7 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.SHADOWPASSWORD], ['08lnuxCM.c36E'])
         self.a.setpassUser(e, 'password', False)
         self.assertEqual(e[libuser.USERPASSWORD], ['x'])
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
         crypted = crypt.crypt('password', e[libuser.SHADOWPASSWORD][0])
         self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
 
@@ -493,8 +600,25 @@ class Tests(unittest.TestCase):
         self.a.setpassUser(e, 'password', False)
         crypted = crypt.crypt('password', e[libuser.USERPASSWORD][0])
         self.assertEqual(e[libuser.USERPASSWORD], [crypted])
+        self.assertEqual(e[libuser.USERPASSWORD][0][:3], '$1$')
         self.assertRaises(KeyError, lambda: e[libuser.SHADOWPASSWORD])
 
+    def testUserSetpass4(self):
+        # ':' in field value
+        e = self.a.initUser('user12_4')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user12_4')
+        self.assert_(':' not in e[libuser.SHADOWPASSWORD][0])
+        self.assertRaises(SystemError, self.a.setpassUser, e, 'a:b', True)
+        self.a.setpassUser(e, 'a:b', False)
+        del e
+        e = self.a.lookupUserByName('user12_4')
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
+        self.assert_(':' not in e[libuser.SHADOWPASSWORD][0])
+        crypted = crypt.crypt('a:b', e[libuser.SHADOWPASSWORD][0])
+        self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
+
     def testUserRemovepass(self):
         e = self.a.initUser('user13_1')
         e[libuser.SHADOWPASSWORD] = '03dgZm5nZvqOc'
@@ -569,6 +693,48 @@ class Tests(unittest.TestCase):
         # Data set up in files_test
         self.assertEqual(self.a.enumerateUsersByGroup('group15_4'), [])
 
+    def testUsersEnumerateByGroupFull1(self):
+        gid = 3401 # Hopefully unique
+        e = self.a.initGroup('group34_1')
+        e[libuser.GIDNUMBER] = gid
+        e[libuser.MEMBERNAME] = 'user34_2'
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initUser('user34_2')
+        e[libuser.GIDNUMBER] = gid + 10
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_1'))
+        v.sort()
+        self.assertEqual(v, [['user34_1'], ['user34_2']])
+
+    def testUsersEnumerateByGroupFull2(self):
+        gid = 3402 # Hopefully unique
+        e = self.a.initGroup('group34_2')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_2'))
+        self.assertEqual(v, [['user34_3']])
+
+    def testUsersEnumerateByGroupFull3(self):
+        gid = 3403 # Hopefully unique
+        e = self.a.initGroup('group34_3')
+        e[libuser.GIDNUMBER] = gid
+        e[libuser.MEMBERNAME] = 'user34_4'
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_4')
+        e[libuser.GIDNUMBER] = gid + 10
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_3'))
+        self.assertEqual(v, [['user34_4']])
+
     def testUsersEnumerateFull1(self):
         e = self.a.initUser('user16_1')
         self.a.addUser(e, False, False)
@@ -690,6 +856,34 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.GROUPNAME], ['group21_3'])
         self.assertEqual(e[libuser.GIDNUMBER], [LARGE_ID + 2130])
 
+    def testGroupAdd4(self):
+        # ':' in field values
+        # libuser.GROUPPASSWORD not tested because lu_shadow_group_add_prep()
+        # always replaces it by 'x'.  libuser.GIDNUMBER not tested because
+        # ent_has_name_and_id() interprets the value as a number.
+        for field in (libuser.GROUPNAME, libuser.SHADOWPASSWORD,
+                      libuser.ADMINISTRATORNAME):
+            e = self.a.initGroup('group_21_4' + field)
+            if e.has_key(field):
+                e[field] = str(e[field][0]) + ':x'
+            else:
+                e[field] = field + ':x'
+            self.assertRaises(RuntimeError, self.a.addGroup, e)
+        e = self.a.initGroup('group21_4' + libuser.MEMBERNAME)
+        e[libuser.MEMBERNAME] = 'group21_4:member'
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group21_4' + libuser.MEMBERNAME)
+        self.assertEqual(e[libuser.MEMBERNAME], ['group21_4:member'])
+
+    def testGroupAdd5(self):
+        # Adding duplicate entries
+        e = self.a.initGroup('group21_5')
+        self.a.addGroup(e)
+        del e
+        e = self.a.initGroup('group21_5')
+        self.assertRaises(RuntimeError, self.a.addGroup, e)
+
     def testGroupMod1(self):
         # A minimal case
         e = self.a.initGroup('group22_1')
@@ -751,6 +945,58 @@ class Tests(unittest.TestCase):
         self.assert_(e)
         self.assertEqual(e[libuser.GIDNUMBER], [LARGE_ID + 2230])
 
+    def testGroupMod4(self):
+        # ':' in field values
+        # libuser.GIDNUMBER not tested because ent_has_name_and_id() interprets
+        # the value as a number.
+        for field in (libuser.GROUPNAME, libuser.GROUPPASSWORD,
+                      libuser.SHADOWPASSWORD, libuser.ADMINISTRATORNAME):
+            e = self.a.initGroup('group22_4' + field)
+            self.a.addGroup(e)
+            del e
+            e = self.a.lookupGroupByName('group22_4' + field)
+            self.assert_(e)
+            if e.has_key(field):
+                self.assert_(':' not in str(e[field][0]))
+                e[field] = str(e[field][0]) + ':x'
+            else:
+                e[field] = field + ':x'
+            self.assertRaises(RuntimeError, self.a.modifyGroup, e)
+        e = self.a.initGroup('group22_4' + libuser.MEMBERNAME)
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group22_4' + libuser.MEMBERNAME)
+        self.assert_(e)
+        not e.has_key(libuser.MEMBERNAME)
+        e[libuser.MEMBERNAME] = 'group22_4:member'
+        self.a.modifyGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group22_4' + libuser.MEMBERNAME)
+        self.assert_(e)
+        self.assertEqual(e[libuser.MEMBERNAME], ['group22_4:member'])
+
+    def testGroupMod5(self):
+        # Nothing to do with the files module: test lu_name_allowed is called.
+        e = self.a.initGroup('group22_5')
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group22_5')
+        e[libuser.GROUPNAME] = 'very_long_name_123456789_123456789_123456789'
+        self.assertRaises(RuntimeError, self.a.modifyGroup, e)
+        del e
+
+    def testGroupMod6(self):
+        # Renaming to create duplicate entries
+        e = self.a.initGroup('group22_6')
+        self.a.addGroup(e)
+        del e
+        e = self.a.initGroup('group22_7')
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group22_7')
+        e[libuser.GROUPNAME] = 'group22_6'
+        self.assertRaises(RuntimeError, self.a.modifyGroup, e)
+
     def testGroupDel(self):
         e = self.a.initGroup('group23_1')
         self.a.addGroup(e)
@@ -890,6 +1136,7 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.SHADOWPASSWORD], ['06aZrb3pzuu/6'])
         self.a.setpassGroup(e, 'password', False)
         self.assertEqual(e[libuser.GROUPPASSWORD], ['x'])
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
         crypted = crypt.crypt('password', e[libuser.SHADOWPASSWORD][0])
         self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
 
@@ -912,6 +1159,7 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.SHADOWPASSWORD], ['07ZZy2Pihe/gg'])
         self.a.setpassGroup(e, 'password', False)
         self.assertEqual(e[libuser.GROUPPASSWORD], ['x'])
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
         crypted = crypt.crypt('password', e[libuser.SHADOWPASSWORD][0])
         self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
 
@@ -923,8 +1171,25 @@ class Tests(unittest.TestCase):
         self.a.setpassGroup(e, 'password', False)
         crypted = crypt.crypt('password', e[libuser.GROUPPASSWORD][0])
         self.assertEqual(e[libuser.GROUPPASSWORD], [crypted])
+        self.assertEqual(e[libuser.GROUPPASSWORD][0][:3], '$1$')
         self.assertRaises(KeyError, lambda: e[libuser.SHADOWPASSWORD])
 
+    def testGroupSetpass4(self):
+        # ':' in field value
+        e = self.a.initGroup('group27_4')
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group27_4')
+        self.assert_(':' not in e[libuser.SHADOWPASSWORD][0])
+        self.assertRaises(SystemError, self.a.setpassGroup, e, 'a:b', True)
+        self.a.setpassGroup(e, 'a:b', False)
+        del e
+        e = self.a.lookupGroupByName('group27_4')
+        self.assertEqual(e[libuser.SHADOWPASSWORD][0][:3], '$1$')
+        self.assert_(':' not in e[libuser.SHADOWPASSWORD][0])
+        crypted = crypt.crypt('a:b', e[libuser.SHADOWPASSWORD][0])
+        self.assertEqual(e[libuser.SHADOWPASSWORD], [crypted])
+
     def testGroupRemovepass(self):
         e = self.a.initGroup('group28_1')
         e[libuser.SHADOWPASSWORD] = '07Js7N.eEhbgs'
@@ -994,6 +1259,48 @@ class Tests(unittest.TestCase):
         # Data set up in files_test
         self.assertEqual(self.a.enumerateGroupsByUser('user30_4'), [])
 
+    def testGroupsEnumerateByUserFull1(self):
+        gid = 3501 # Hopefully unique
+        e = self.a.initUser('user35_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        e = self.a.initGroup('group35_2')
+        e[libuser.GIDNUMBER] = gid + 10
+        e[libuser.MEMBERNAME] = 'user35_1'
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_1'))
+        v.sort()
+        self.assertEqual(v, [['group35_1'], ['group35_2']])
+
+    def testGroupsEnumerateByUserFull2(self):
+        gid = 3502 # Hopefully unique
+        e = self.a.initUser('user35_2')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_2'))
+        self.assertEqual(v, [['group35_3']])
+
+    def testGroupsEnumerateByUserFull3(self):
+        gid = 3503 # Hopefully unique
+        e = self.a.initUser('user35_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_4')
+        e[libuser.GIDNUMBER] = gid + 10
+        e[libuser.MEMBERNAME] = 'user35_3'
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_3'))
+        self.assertEqual(v, [['group35_4']])
+
     def testGroupsEnumerateFull1(self):
         e = self.a.initGroup('group31_1')
         self.a.addGroup(e)
@@ -1016,6 +1323,26 @@ class Tests(unittest.TestCase):
         self.a.addGroup(e)
         self.assertEqual(self.a.enumerateGroupsFull('group31_3:*'), [])
 
+    # ValidateIdValue is unrelated to the files module.
+    def testValidateIdValue(self):
+        libuser.validateIdValue(0)
+        libuser.validateIdValue(1)
+        libuser.validateIdValue(500)
+        libuser.validateIdValue(500L)
+        self.assertRaises(TypeError, libuser.validateIdValue, 'abc')
+        # OverflowError if id_t is unsigned, ValueError otherwise
+        self.assertRaises((ValueError, OverflowError), libuser.validateIdValue,
+                          -1)
+        self.assertRaises((ValueError, OverflowError), libuser.validateIdValue,
+                          -2)
+        self.assertRaises(ValueError, libuser.validateIdValue,
+                          libuser.VALUE_INVALID_ID)
+        if libuser.VALUE_INVALID_ID > 0:
+            libuser.validateIdValue(libuser.VALUE_INVALID_ID - 1)
+            if libuser.VALUE_INVALID_ID < 2 ** 64 - 1:
+                self.assertRaises(OverflowError, libuser.validateIdValue,
+                                  2 ** 64 - 1)
+
     def tearDown(self):
         del self.a
 
