@@ -65,6 +65,7 @@ class Tests(unittest.TestCase):
         e = self.a.lookupUserByName('user6_1')
         self.assert_(e)
         self.assertEqual(e[libuser.USERNAME], ['user6_1'])
+        self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}!!'])
 
     def testUserAdd2(self):
         # A maximal case
@@ -139,6 +140,68 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.UIDNUMBER], [LARGE_ID + 640])
         self.assertEqual(e[libuser.GIDNUMBER], [LARGE_ID + 641])
 
+    def testUserAdd5(self):
+        # Let the user specify dangerous/dubious home directory paths explicitly
+        for name in ('.', '..'):
+            e = self.a.initUser(name)
+            e[libuser.HOMEDIRECTORY] = '/home/' + name
+            self.a.addUser(e, False, False)
+            del e
+            e = self.a.lookupUserByName(name)
+            self.assertEqual(e[libuser.HOMEDIRECTORY], ['/home/' + name])
+            self.a.deleteUser(e, False, False)
+        e = self.a.initUser('user6_5')
+        e[libuser.HOMEDIRECTORY] = '/home/..'
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_5')
+        self.assertEqual(e[libuser.HOMEDIRECTORY], ['/home/..'])
+        # ... but nevere create such home directory paths by default
+        for name in ('.', '..'):
+            e = self.a.initUser(name)
+            self.assertRaises(RuntimeError, self.a.addUser, e, False, False)
+
+    def testUserAdd6(self):
+        # Adding duplicate entries
+        e = self.a.initUser('user6_6')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.initUser('user6_6')
+        self.assertRaises(RuntimeError, self.a.addUser, e, False, False)
+
+    def testUserAdd7(self):
+        # Adding commonName if it is not defined:
+        # - Explicitly set
+        e = self.a.initUser('user6_7')
+        e[libuser.COMMONNAME] = 'Common Name'
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_7')
+        self.assertEqual(e[libuser.COMMONNAME], ['Common Name'])
+        # - Defaulted from GECOS
+        e = self.a.initUser('user6_8')
+        e[libuser.GECOS] = 'Full Name,Office,1234,4321'
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_8')
+        self.assertEqual(e[libuser.COMMONNAME], ['Full Name'])
+        # Defaulted from user name
+        e = self.a.initUser('user6_9')
+        self.assertEqual(e[libuser.GECOS], ['user6_9'])
+        e.clear(libuser.GECOS)
+        self.assertRaises(KeyError, lambda x: x[libuser.GECOS], e)
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_9')
+        self.assertEqual(e[libuser.COMMONNAME], ['user6_9'])
+        # Defaulted from user name if GECOS is empty
+        e = self.a.initUser('user6_10')
+        e[libuser.GECOS] = ''
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user6_10')
+        self.assertEqual(e[libuser.COMMONNAME], ['user6_10'])
+
     def testUserMod1(self):
         # A minimal case
         e = self.a.initUser('user7_1')
@@ -237,6 +300,18 @@ class Tests(unittest.TestCase):
         e = self.a.lookupUserByName('user7_4')
         self.assert_(e)
 
+    def testUserMod5(self):
+        # Renaming to create duplicate entries
+        e = self.a.initUser('user7_5')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.initUser('user7_6')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user7_6')
+        e[libuser.USERNAME] = 'user7_5'
+        self.assertRaises(RuntimeError, self.a.modifyUser, e, False)
+
     def testUserDel(self):
         e = self.a.initUser('user8_1')
         self.a.addUser(e, False, False)
@@ -268,6 +343,16 @@ class Tests(unittest.TestCase):
         self.a.lockUser(e)
         self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}!00as1wm0AZG56'])
 
+    def testUserLock3(self):
+        e = self.a.initUser('user9_3')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user9_3')
+        self.assertEqual(e[libuser.USERPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.lockUser, e)
+
     def testUserUnlock1(self):
         e = self.a.initUser('user10_1')
         self.a.addUser(e, False, False)
@@ -298,6 +383,16 @@ class Tests(unittest.TestCase):
         self.a.unlockUser(e)
         self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}'])
 
+    def testUserUnlock4(self):
+        e = self.a.initUser('user10_4')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user10_4')
+        self.assertEqual(e[libuser.USERPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.unlockUser, e)
+
     def testUserUnlockNonempty1(self):
         e = self.a.initUser('user32_1')
         self.a.addUser(e, False, False)
@@ -330,6 +425,16 @@ class Tests(unittest.TestCase):
         e = self.a.lookupUserByName('user32_3')
         self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}!'])
 
+    def testUserUnlockNonempty4(self):
+        e = self.a.initUser('user32_4')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user32_4')
+        self.assertEqual(e[libuser.USERPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.unlockUser, e, True)
+
     def testUserIslocked1(self):
         e = self.a.initUser('user11_1')
         self.a.addUser(e, False, False)
@@ -360,6 +465,7 @@ class Tests(unittest.TestCase):
         self.a.setpassUser(e, 'password', False)
         del e
         e = self.a.lookupUserByName('user12_1')
+        self.assertNotEqual(e[libuser.USERPASSWORD][0][7], '$')
         crypted = crypt.crypt('password', e[libuser.USERPASSWORD][0][7:])
         self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}' + crypted])
         self.assert_(e[libuser.SHADOWLASTCHANGE][0] > 10000)
@@ -377,6 +483,7 @@ class Tests(unittest.TestCase):
         self.a.setpassUser(e, 'password', False)
         v = e[libuser.USERPASSWORD]
         v.sort()
+        self.assertNotEqual(v[1][7], '$')
         crypted = crypt.crypt('password', v[1][7:])
         self.assertEqual(v, ['unknown', '{CRYPT}' + crypted])
 
@@ -390,6 +497,36 @@ class Tests(unittest.TestCase):
         v.sort()
         self.assertEqual(v, ['unknown1', 'unknown2'])
         self.a.setpassUser(e, 'password', False)
+        self.assertEqual(e[libuser.USERPASSWORD][0][7:10], '$1$')
+        crypted = crypt.crypt('password', e[libuser.USERPASSWORD][0][7:])
+        self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}' + crypted])
+
+    def testUserSetpass4(self):
+        # Test that crypt_style is honored on accounts with no existing hash
+        e = self.a.initUser('user12_4')
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user12_4')
+        self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}!!'])
+        self.a.setpassUser(e, 'password', False)
+        del e
+        e = self.a.lookupUserByName('user12_4')
+        self.assertEqual(e[libuser.USERPASSWORD][0][7:10], '$1$')
+        crypted = crypt.crypt('password', e[libuser.USERPASSWORD][0][7:])
+        self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}' + crypted])
+
+    def testUserSetpass5(self):
+        # Test that existing hash is recognized on locked accounts
+        e = self.a.initUser('user12_5')
+        e[libuser.USERPASSWORD] = ['{CRYPT}!05q.yCrQKlZy6']
+        self.a.addUser(e, False, False)
+        del e
+        e = self.a.lookupUserByName('user12_5')
+        self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}!05q.yCrQKlZy6'])
+        self.a.setpassUser(e, 'password', False)
+        del e
+        e = self.a.lookupUserByName('user12_5')
+        self.assertNotEqual(e[libuser.USERPASSWORD][0][7], '$')
         crypted = crypt.crypt('password', e[libuser.USERPASSWORD][0][7:])
         self.assertEqual(e[libuser.USERPASSWORD], ['{CRYPT}' + crypted])
 
@@ -493,6 +630,48 @@ class Tests(unittest.TestCase):
         v.sort()
         self.assertEqual(v, [['user16_1'], ['user16_2']])
 
+    def testUsersEnumerateByGroupFull1(self):
+        gid = 3401 # Hopefully unique
+        e = self.a.initGroup('group34_1')
+        e[libuser.GIDNUMBER] = gid
+        e[libuser.MEMBERNAME] = 'user34_2'
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initUser('user34_2')
+        e[libuser.GIDNUMBER] = gid + 10
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_1'))
+        v.sort()
+        self.assertEqual(v, [['user34_1'], ['user34_2']])
+
+    def testUsersEnumerateByGroupFull2(self):
+        gid = 3402 # Hopefully unique
+        e = self.a.initGroup('group34_2')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_2'))
+        self.assertEqual(v, [['user34_3']])
+
+    def testUsersEnumerateByGroupFull3(self):
+        gid = 3403 # Hopefully unique
+        e = self.a.initGroup('group34_3')
+        e[libuser.GIDNUMBER] = gid
+        e[libuser.MEMBERNAME] = 'user34_4'
+        self.a.addGroup(e)
+        e = self.a.initUser('user34_4')
+        e[libuser.GIDNUMBER] = gid + 10
+        self.a.addUser(e, False, False)
+        v = map(lambda x: x[libuser.USERNAME],
+                self.a.enumerateUsersByGroupFull('group34_3'))
+        self.assertEqual(v, [['user34_4']])
+
     def testGroupLookupName(self):
         e = self.a.initGroup('group17_1')
         self.a.addGroup(e)
@@ -533,6 +712,7 @@ class Tests(unittest.TestCase):
         e = self.a.lookupGroupByName('group21_1')
         self.assert_(e)
         self.assertEqual(e[libuser.GROUPNAME], ['group21_1'])
+        self.assertRaises(KeyError, lambda x: x[libuser.GROUPPASSWORD], e)
 
     def testGroupAdd2(self):
         # A maximal case
@@ -563,6 +743,14 @@ class Tests(unittest.TestCase):
         self.assertEqual(e[libuser.GROUPNAME], ['group21_3'])
         self.assertEqual(e[libuser.GIDNUMBER], [LARGE_ID + 2130])
 
+    def testGroupAdd4(self):
+        # Adding duplicate entries
+        e = self.a.initGroup('group21_4')
+        self.a.addGroup(e)
+        del e
+        e = self.a.initGroup('group21_4')
+        self.assertRaises(RuntimeError, self.a.addGroup, e)
+
     def testGroupMod1(self):
         # A minimal case
         e = self.a.initGroup('group22_1')
@@ -618,6 +806,18 @@ class Tests(unittest.TestCase):
         self.assert_(e)
         self.assertEqual(e[libuser.GIDNUMBER], [LARGE_ID + 2230])
 
+    def testGroupMod4(self):
+        # Renaming to create duplicate entries
+        e = self.a.initGroup('group22_4')
+        self.a.addGroup(e)
+        del e
+        e = self.a.initGroup('group22_5')
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group22_5')
+        e[libuser.GROUPNAME] = 'group22_4'
+        self.assertRaises(RuntimeError, self.a.modifyGroup, e)
+
     def testGroupDel(self):
         e = self.a.initGroup('group23_1')
         self.a.addGroup(e)
@@ -649,6 +849,16 @@ class Tests(unittest.TestCase):
         self.a.lockGroup(e)
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}!04cmES7HM6wtg'])
 
+    def testGroupLock3(self):
+        e = self.a.initGroup('group24_3')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group24_3')
+        self.assertEqual(e[libuser.GROUPPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.lockGroup, e)
+
     def testGroupUnlock1(self):
         e = self.a.initGroup('group25_1')
         self.a.addGroup(e)
@@ -679,6 +889,16 @@ class Tests(unittest.TestCase):
         self.a.unlockGroup(e)
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}'])
 
+    def testGroupUnlock4(self):
+        e = self.a.initGroup('group25_4')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group25_4')
+        self.assertEqual(e[libuser.GROUPPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.unlockGroup, e)
+
     def testGroupUnlockNonempty1(self):
         e = self.a.initGroup('group33_1')
         self.a.addGroup(e)
@@ -711,6 +931,16 @@ class Tests(unittest.TestCase):
         e = self.a.lookupGroupByName('group33_3')
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}!'])
 
+    def testGroupUnlockNonempty4(self):
+        e = self.a.initGroup('group33_4')
+        e[libuser.USERPASSWORD] = '{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group33_4')
+        self.assertEqual(e[libuser.GROUPPASSWORD],
+                         ['{MD5}Xr4ilOzQ4PCOq3aQ0qbuaQ=='])
+        self.assertRaises(RuntimeError, self.a.unlockGroup, e, True)
+
     def testGroupIsLocked1(self):
         e = self.a.initGroup('group26_1')
         self.a.addGroup(e)
@@ -737,6 +967,7 @@ class Tests(unittest.TestCase):
         e = self.a.lookupGroupByName('group27_1')
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}06aZrb3pzuu/6'])
         self.a.setpassGroup(e, 'password', False)
+        self.assertNotEqual(e[libuser.GROUPPASSWORD][0][7], '$')
         crypted = crypt.crypt('password', e[libuser.GROUPPASSWORD][0][7:])
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}' + crypted])
 
@@ -752,6 +983,7 @@ class Tests(unittest.TestCase):
         self.a.setpassGroup(e, 'password', False)
         v = e[libuser.GROUPPASSWORD]
         v.sort()
+        self.assertNotEqual(e[libuser.GROUPPASSWORD][1][7], '$')
         crypted = crypt.crypt('password', v[1][7:])
         self.assertEqual(v, ['unknown', '{CRYPT}' + crypted])
 
@@ -765,6 +997,36 @@ class Tests(unittest.TestCase):
         v.sort()
         self.assertEqual(v, ['unknown1', 'unknown2'])
         self.a.setpassGroup(e, 'password', False)
+        self.assertEqual(e[libuser.GROUPPASSWORD][0][7:10], '$1$')
+        crypted = crypt.crypt('password', e[libuser.GROUPPASSWORD][0][7:])
+        self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}' + crypted])
+
+    def testGroupSetpass4(self):
+        # Test that crypt_style is honored on accounts with no existing hash
+        e = self.a.initGroup('group27_4')
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group27_4')
+        self.assertRaises(KeyError, lambda x: x[libuser.GROUPPASSWORD], e)
+        self.a.setpassGroup(e, 'password', False)
+        del e
+        e = self.a.lookupGroupByName('group27_4')
+        self.assertEqual(e[libuser.GROUPPASSWORD][0][7:10], '$1$')
+        crypted = crypt.crypt('password', e[libuser.GROUPPASSWORD][0][7:])
+        self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}' + crypted])
+
+    def testGroupSetpass5(self):
+        # Test that crypt_style is honored on accounts with no existing hash
+        e = self.a.initGroup('group27_5')
+        e[libuser.GROUPPASSWORD] = ['{CRYPT}!05XMoJIk1se52']
+        self.a.addGroup(e)
+        del e
+        e = self.a.lookupGroupByName('group27_5')
+        self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}!05XMoJIk1se52'])
+        self.a.setpassGroup(e, 'password', False)
+        del e
+        e = self.a.lookupGroupByName('group27_5')
+        self.assertNotEqual(e[libuser.GROUPPASSWORD][0][7], '$')
         crypted = crypt.crypt('password', e[libuser.GROUPPASSWORD][0][7:])
         self.assertEqual(e[libuser.GROUPPASSWORD], ['{CRYPT}' + crypted])
 
@@ -862,9 +1124,51 @@ class Tests(unittest.TestCase):
         v.sort()
         self.assertEqual(v, [['group31_1'], ['group31_2']])
 
+    def testGroupsEnumerateByUserFull1(self):
+        gid = 3501 # Hopefully unique
+        e = self.a.initUser('user35_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_1')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        e = self.a.initGroup('group35_2')
+        e[libuser.GIDNUMBER] = gid + 10
+        e[libuser.MEMBERNAME] = 'user35_1'
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_1'))
+        v.sort()
+        self.assertEqual(v, [['group35_1'], ['group35_2']])
+
+    def testGroupsEnumerateByUserFull2(self):
+        gid = 3502 # Hopefully unique
+        e = self.a.initUser('user35_2')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_2'))
+        self.assertEqual(v, [['group35_3']])
+
+    def testGroupsEnumerateByUserFull3(self):
+        gid = 3503 # Hopefully unique
+        e = self.a.initUser('user35_3')
+        e[libuser.GIDNUMBER] = gid
+        self.a.addUser(e, False, False)
+        e = self.a.initGroup('group35_4')
+        e[libuser.GIDNUMBER] = gid + 10
+        e[libuser.MEMBERNAME] = 'user35_3'
+        self.a.addGroup(e)
+        v = map(lambda x: x[libuser.GROUPNAME],
+                self.a.enumerateGroupsByUserFull('user35_3'))
+        self.assertEqual(v, [['group35_4']])
+
     def tearDown(self):
-        del self.a        
-        
+        del self.a
+
 
 if __name__ == '__main__':
     unittest.main()
